/*
 * Copyright (c) 2015 Marcus Lundblad
 *
 * GNOME Radio is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * GNOME Radio is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with GNOME Radio; if not, see <http://www.gnu.org/licenses/>
 *
 * Author: Marcus Lundblad <ml@update.uu.se>
 */

#include "radio-osm-relation.h"

struct _RadioOSMRelationPrivate
{
  GList *members;
};

G_DEFINE_TYPE_WITH_PRIVATE (RadioOSMRelation, radio_osm_relation,
                            RADIO_TYPE_OSMOBJECT);

typedef struct
{
  char *role;
  guint type;
  guint64 ref;
} RadioOSMRelationMember;

static void
radio_osm_relation_member_free (gpointer data)
{
  RadioOSMRelationMember *member = (RadioOSMRelationMember *) data;

  g_free (member->role);
}

static void
radio_osm_relation_dispose (GObject *object)
{
  RadioOSMRelation *relation = RADIO_OSMRELATION (object);

  g_list_free_full (relation->priv->members, radio_osm_relation_member_free);
  relation->priv->members = NULL;

  G_OBJECT_CLASS (radio_osm_relation_parent_class)->dispose (object);
}

static const char *
radio_osm_relation_get_xml_tag_name (void)
{
  return "relation";
}

static const char *
radio_osm_relation_member_type_to_string (guint type)
{
  switch (type) {
  case MEMBER_TYPE_NODE:
    return "node";
  case MEMBER_TYPE_WAY:
    return "way";
  case MEMBER_TYPE_RELATION:
    return "relation";
  default:
    g_warning ("Unknown relation member type: %d\n", type);
    return NULL;
  }
}

static xmlNodePtr
radio_osm_relation_get_member_node (const RadioOSMRelationMember *member)
{
  xmlNodePtr node = xmlNewNode (NULL, (xmlChar *) "member");
  char buf[16];
  
  if (member->role)
    xmlNewProp (node, (xmlChar *) "role", (xmlChar *) g_strdup (member->role));

  xmlNewProp (node, (xmlChar *) "type",
              (xmlChar *) radio_osm_relation_member_type_to_string (member->type));
  g_snprintf (buf, 16, "%" G_GUINT64_FORMAT, member->ref);
  xmlNewProp (node, (xmlChar *) "ref", (xmlChar *) buf);

  return node;
}

static xmlNodePtr
radio_osm_relation_get_xml_child_nodes (const RadioOSMObject *object)
{
  RadioOSMRelation *relation = RADIO_OSMRELATION ((RadioOSMObject *) object);
  xmlNodePtr nodes = NULL;
  const GList *members = relation->priv->members;
  
  if (members)
    {
      const GList *iter;
      nodes = radio_osm_relation_get_member_node ((RadioOSMRelationMember *)
                                                 members->data);

      for (iter = members->next; iter; iter = iter->next)
        {
          xmlAddSibling (nodes, radio_osm_relation_get_member_node (
            (RadioOSMRelationMember *) iter->data));
        }
    }

  return nodes;
}

static void
radio_osm_relation_class_init (RadioOSMRelationClass *klass)
{
  GObjectClass *relation_class = G_OBJECT_CLASS (klass);
  RadioOSMObjectClass *object_class = RADIO_OSMOBJECT_CLASS (klass);

  relation_class->dispose = radio_osm_relation_dispose;
  object_class->get_xml_tag_name = radio_osm_relation_get_xml_tag_name;
  object_class->get_xml_child_nodes = radio_osm_relation_get_xml_child_nodes;
}

static void
radio_osm_relation_init (RadioOSMRelation *relation)
{
  relation->priv = radio_osm_relation_get_instance_private (relation);
}

RadioOSMRelation *
radio_osm_relation_new (guint64 id, guint version, guint64 changeset)
{
  return g_object_new (RADIO_TYPE_OSMRELATION,
                       "id", id,
                       "version", version,
                       "changeset", changeset, NULL);
}

void
radio_osm_relation_add_member (RadioOSMRelation *relation, const gchar *role,
                              guint type, guint64 ref)
{
  RadioOSMRelationMember *member = g_new (RadioOSMRelationMember, 1);

  member->role = g_strdup (role);
  member->type = type;
  member->ref = ref;

  relation->priv->members = g_list_append (relation->priv->members, member);
}

